/**********************************************************************
 *
 * ntimeset.c
 * version 1.1
 *
 * ntimeset allows you to set the time on a Notes/Domino server to
 * match the time that is set on the operating system. It uses an
 * unsupported/undocumented function (OSSetCurrentDateTime), so you
 * should be careful using this program, especially while other Notes
 * processes are running.
 *
 * To compile this program, you need to have your compiler all set up
 * to use the Notes C-API libraries (I think I used the 5.06 libraries,
 * but other versions will probably work too, as long as they're not
 * too old). You should put the final executable in the Notes program
 * directory of the server you intend to run it on. I have only tested
 * compiling this on Windows...other operating systems may require some
 * tweaks to the code.
 *
 * One interesting aspect of this code is that it uses an unsupported
 * function (OSSetCurrentDateTime) that doesn't exist in any of the
 * library files that ship with the Notes API toolkits. This program
 * shows you how to get a handle to the function directly from the
 * nnotes.dll file using LoadLibrary and GetProcAddress.
 *
 * At the time of this writing, the current version of the source code 
 * is still available at the http://www.nsftools.com site (in the "Tools"
 * section), although my goal is to turn this into an OpenNTF project so
 * that other people can contribute to the codestream. Check there too:
 * http://www.OpenNTF.org (or http://www.notesoss.org for the oldtimers).
 *
 * Written by Julian Robichaux
 * http://www.nsftools.com
 *
 * You may use this code in any way you'd like, as long as you don't
 * hold me liable for anything, and you don't tell people that you 
 * wrote it yourself. If you're feeling really generous, please include
 * my name and website whenever you use this code in your programs.
 * Thanks!
 *
 * HISTORY
 * version 1.1   - first open source version of the program
 *
 **********************************************************************/


/* standard C libraries */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

/* Notes API libraries */
#include <global.h>
#include <nsfdb.h>
#include <ostime.h>
#include <osmisc.h>
#include <misc.h>


// OSSetCurrentDateTime parameter structure (we'll need this
// when we use GetProcAddress to get a handle to the function
typedef STATUS (LNPUBLIC *LPSetTime)(TIMEDATE far *);


// globals (set when we check the command line parameters)
int displayTimes = FALSE;
int setCurrentTime = FALSE;


// forward declarations, so we don't have to write a header file
int ProcessArgs (int argc, char *argv[]);
void DisplayTime ();
void SetTime ();
STATUS GetStdTime (TIMEDATE *nTime);
void PrintUsage ();
void PrintError (STATUS nError, char *otherText);


/*********************************************************************************
 * The NotesMain function is the main entry point for Notes API programs, and
 * therefore it's the main entry point in this program as well
 *********************************************************************************/
STATUS LNPUBLIC NotesMain (int argc, char *argv[])
{
	STATUS nError = NOERROR;

	// check the command line arguments, and exit early if there
	// was a problem
	if (!ProcessArgs(argc, argv))
		return (NOERROR);

	// just display the help if we're not supposed to display the
	// current times or set the times
	if ((!displayTimes) && (!setCurrentTime))
	{
		PrintUsage();
		return (NOERROR);
	}

	// display the times, if that's what the user wants...
	if (displayTimes)
		DisplayTime();

	// ...and set it if we're supposed to do that
	if (setCurrentTime)
		SetTime();

	// bye bye
	return NOERROR;
}


/*********************************************************************************
 * This will display the OS time and the Notes server time to stdout
 *********************************************************************************/
void DisplayTime ()
{
	TIMEDATE notesTime, osTime;
	char timeText[MAXALPHATIMEDATE + 1];
	WORD textLen;
	STATUS nError = NOERROR;

	/* get the Notes time */
	OSCurrentTIMEDATE (&notesTime);
	nError = ConvertTIMEDATEToText(NULL, NULL, &notesTime, timeText,
					MAXALPHATIMEDATE, &textLen);
	timeText[textLen] = '\0';

	if (nError)
		PrintError(nError, "Error converting Notes Time.");
	else
		printf("Notes time is:  %s\n", timeText);

	/* get the OS time */
	nError = GetStdTime (&osTime);
	if (nError)
	{
		printf("Error getting OS time.\n");
	}  else  {
		nError = ConvertTIMEDATEToText(NULL, NULL, &osTime, timeText,
						MAXALPHATIMEDATE, &textLen);
		timeText[textLen] = '\0';

		if (nError)
			PrintError(nError, "Error converting OS Time.");
		else
			printf("OS time is:     %s\n", timeText);
	}

	return;
	
}


/*********************************************************************************
 * This function actually sets the time on the Notes/Domino server to match
 * the time of the operating system. We do this using the OSSetCurrentDateTime
 * function that's exported in the main notes library file (nnotes.dll on a
 * Windows machine); however, since it's not in the API library files, we have
 * to use LoadLibrary and GetProcAddress to get a handle to the function before
 * we can call it.
 *********************************************************************************/
void SetTime ()
{
	HMODULE hNotesDLL = 0;
	LPSetTime OSSetTime;
	TIMEDATE osTime;
	char timeText[MAXALPHATIMEDATE + 1];
	WORD textLen;
	STATUS nError = NOERROR;

	/* try to get a handle to nnotes.dll, and the OSSetCurrentDateTime function
	   (because this is an undocumented function and it's not part of the standard
	   Notes API libraries, we'll need to use LoadLibrary to read it from the 
	   nnotes.dll file -- this file will have a different name on non-Windows OSes)*/
	hNotesDLL = LoadLibrary("nnotes.dll");
	if (hNotesDLL)
		OSSetTime = (LPSetTime) GetProcAddress(hNotesDLL, "OSSetCurrentDateTime");

	/* if we couldn't get a handle to the exported function, exit early */
	if (!OSSetTime)
	{
		printf("Error: Set Time function not found\n");
		goto cleanUp;
	}

	/* get the OS time */
	nError = GetStdTime (&osTime);
	if (nError)
	{
		printf("Error getting OS time.\n");
		goto cleanUp;
	}

	/* get the time string for later */
	nError = ConvertTIMEDATEToText(NULL, NULL, &osTime, timeText,
					MAXALPHATIMEDATE, &textLen);
	timeText[textLen] = '\0';
	//printf("Current OS time is %s\n", timeText);

	/* attempt to set the time -- I had to disable optimizations in the 
	   compiler for this to work in Release version... */
	nError = OSSetTime(&osTime);
	if (nError)
		PrintError(nError, "Error setting Notes time.");
	else
		printf("Notes time successfully set to %s\n", timeText);


cleanUp:
	if (hNotesDLL)
		FreeLibrary(hNotesDLL);

	return;
}


/*********************************************************************************
 * Convert the current OS time (using standard C-library functions) to a Notes
 * TIMEDATE structure -- if we use OSCurrentTIMEDATE, we get the Notes server
 * time, not the OS time (ironically)
 *********************************************************************************/
STATUS GetStdTime (TIMEDATE *nTime)
{
	static TIME tt;
	time_t cTime;
	struct tm * cTimeStruct;
	STATUS nError = NOERROR;

	// get the OS time as a time_t structure
	time ( &cTime );
	cTimeStruct = localtime ( &cTime );

	// convert it to a TIMEDATE structure by manually setting the date,
	// time zone, hour, minute, and seconds
	OSCurrentTIMEDATE ( &tt.GM );
	OSCurrentTimeZone ( &tt.zone, &tt.dst );
	TimeGMToLocalZone ( &tt );
	tt.hour = cTimeStruct->tm_hour;
	tt.minute = cTimeStruct->tm_min;
	tt.second = cTimeStruct->tm_sec;
	tt.hundredth = 0;
	nError = TimeLocalToGM ( &tt );

	// and return the TIMEDATE that we just created
	*nTime = tt.GM;
	return nError;
}


/*********************************************************************************
 * Get the command line arguments that were passed to this program, and set
 * the global variables accordingly.
 *********************************************************************************/
int ProcessArgs (int argc, char *argv[])
{
	int i;

	// if no arguments were passed, just return FALSE
	if (argc < 2)
	{
		PrintUsage();
		return FALSE;
	}

	// we'll accept both "/" and "-" as parameter prefixes
	for (i = 1; i < argc; i++)
	{
		if ((argv[i][0] == '-') | (argv[i][0] == '/'))
		{
			switch (argv[i][1])
			{
			case '?' :
				// if any of the arguments were "?", just display the
				// usage instructions for this program and return FALSE
				PrintUsage();
				return FALSE;

			case 'd' :
			case 'D' :
				// the "D" parameter means we should display the OS time
				// and the Notes/Domino server time to the console
				displayTimes = TRUE;
				break;

			case 's' :
			case 'S' :
				// the "S" parameter means we should set the Notes/Domino
				// server time to match the OS time
				setCurrentTime = TRUE;
				displayTimes = TRUE;
				break;

			default :
				break;
			}
		}

	}

	// if we got this far, carry on...
	return TRUE;

}


/*********************************************************************************
 * Tell the user how to use this program
 *********************************************************************************/
void PrintUsage ()
{
	printf("Sets the time on the local Notes server or client equal to the OS time.\n\n");
	printf("USAGE: ntimeset [/d] [/s]\n\n");
	printf("  /d      displays the current Notes time and OS time\n");
	printf("  /s      sets the Notes time to match the current OS time\n");
	printf("\ncopyright 2003, Julian Robichaux -- nsftools.com\nversion 1.1\n");
}


/*********************************************************************************
 * Convenience function for displaying the text of an error that occurred
 *********************************************************************************/
void PrintError (STATUS nError, char *otherText)
{
#define ERR_TEXT   256

	char errorString[ERR_TEXT];

	if (OSLoadString(NULLHANDLE, ERR(nError), errorString, ERR_TEXT-1))
		printf("%s (Error: %s)\n", otherText, errorString);
}


